#include <stdio.h>
#include <stdlib.h>
#include "sparse_matrix.h"
#include "matrix_operation.h"

int read_dims_ignore_comment(SparseMatrix *matrix, FILE *file) {
    char buffer[1024];
    int read_dims = 0;
    while (fgets(buffer, sizeof(buffer), file) != NULL) {
        char *line = buffer;
        if (*line == '%' || *line == '\n' || *line == '\0') {
            continue;
        } else {
            if (sscanf(line, "%d %d", &matrix->num_nodes, &matrix->num_arcs) != 2) {
                return 1;
            }
            read_dims = 1;
            break;
        }
    }
    return (!read_dims);
}

void parse_arcs(SparseMatrix *matrix, FILE *file) {
    for (int i = 0; i < matrix->num_arcs; ++i) {
        char line[256];
        if (fgets(line, sizeof(line), file) == NULL) {
            fprintf(stderr, "Failed to read arc %d\n", i);
            exit(5);
        }
        char *ptr = line;
        int origin, dest;
        double value = 1.0;
        if (sscanf(ptr, "%d %d %lf", &origin, &dest, &value) < 2) {
            fprintf(stderr, "Failed to read arc %d\n", i);
            exit(5);
        } else {
            matrix->arcs[i].origin = origin;
            matrix->arcs[i].dest = dest;
            matrix->arcs[i].value = value;
        }
    }
}

SparseMatrix* read_sparse_matrix_from_mtx(const char *filename) {
    FILE *file = fopen(filename, "r");
    if (!file) {
        fprintf(stderr, "Failed to read file\n");
        exit(1);
    }

    SparseMatrix *matrix = (SparseMatrix*)malloc(sizeof(SparseMatrix));
    if (!matrix) {
        fprintf(stderr, "Failed to allocate memory for matrix\n");
        exit(3);
    }

    if (read_dims_ignore_comment(matrix, file)) {
        fprintf(stderr, "Failed to read matrix dimensions\n");
    }

    matrix->arcs = (Arc*)malloc(matrix->num_arcs * sizeof(Arc));
    if (!matrix->arcs) {
        fprintf(stderr, "Failed to allocate memory for arcs\n");
        exit(4);
    }

    parse_arcs(matrix, file);

    // Sort arcs by origin
    qsort(matrix->arcs, matrix->num_arcs, sizeof(Arc), compare_arcs);
    // Build row_ptr
    matrix->row_ptr = (int *)malloc((matrix->num_nodes + 1) * sizeof(int));
    int current_origin = 0;
    matrix->row_ptr[0] = 0;
    for (int k = 0; k < matrix->num_arcs; k++) {
        while (matrix->arcs[k].origin > current_origin) {
            current_origin++;
            matrix->row_ptr[current_origin] = k;
        }
    }
    for (int i = current_origin + 1; i <= matrix->num_nodes; i++) {
        matrix->row_ptr[i] = matrix->num_arcs;
    }

    fclose(file);
    return matrix;
}
