#include <stdio.h>
#include <stdlib.h>
#include "read_from_rb.h"
#include "sparse_matrix.h"
#include "matrix_operation.h"

SparseMatrix* read_sparse_matrix_from_rb(const char *filename) {
    FILE *file = fopen(filename, "r");
    if (!file) {
        fprintf(stderr, "Failed to open file %s\n", filename);
        exit(1);
    }

    SparseMatrix *matrix = (SparseMatrix*)malloc(sizeof(SparseMatrix));
    if (!matrix) {
        fprintf(stderr, "Failed to allocate memory for matrix\n");
        exit(2);
    }

    char buffer[256];
    fgets(buffer, sizeof(buffer), file);
    fgets(buffer, sizeof(buffer), file);

    fgets(buffer, sizeof(buffer), file);
    char type[4];
    int rows, cols, nonzero;
    if (sscanf(buffer, "%3s %d %d %d", type, &rows, &cols, &nonzero) != 4) {
        fprintf(stderr, "Failed to read matrix metadata\n");
        exit(3);
    }

    matrix->num_nodes = rows;
    matrix->num_arcs = nonzero;
    matrix->arcs = (Arc*)malloc(nonzero * sizeof(Arc));
    if (!matrix->arcs) {
        fprintf(stderr, "Failed to allocate memory for arcs\n");
        exit(4);
    }

    fgets(buffer, sizeof(buffer), file);

    int *col_ptr = (int*)malloc((cols + 1) * sizeof(int));
    int ptr_index = 0;
    while (ptr_index < cols + 1) {
        if (fgets(buffer, sizeof(buffer), file) == NULL) {
            fprintf(stderr, "Unexpected end of file while reading column pointers\n");
            exit(5);
        }
        char *ptr = buffer;
        int num_read;
        while (sscanf(ptr, "%d%n", &col_ptr[ptr_index], &num_read) == 1) {
            ptr += num_read;
            ptr_index++;
        }
    }

    int *row_ind = (int*)malloc(nonzero * sizeof(int));
    int row_index = 0;
    while (row_index < nonzero) {
        if (fgets(buffer, sizeof(buffer), file) == NULL) {
            fprintf(stderr, "Unexpected end of file while reading row indices\n");
            exit(6);
        }
        char *ptr = buffer;
        int num_read;
        while (sscanf(ptr, "%d%n", &row_ind[row_index], &num_read) == 1) {
            ptr += num_read;
            row_index++;
        }
    }

    double *values = (double*)malloc(nonzero * sizeof(double));
    int val_index = 0;
    while (val_index < nonzero) {
        if (fgets(buffer, sizeof(buffer), file) == NULL) {
            fprintf(stderr, "Unexpected end of file while reading values\n");
            exit(7);
        }
        char *ptr = buffer;
        int num_read;
        double val;
        while (sscanf(ptr, "%lf%n", &val, &num_read) == 1) {
            values[val_index] = val;
            ptr += num_read;
            val_index++;
        }
    }

    int arc_idx = 0;
    for (int j = 0; j < cols; j++) {
        int start = col_ptr[j] - 1;
        int end = col_ptr[j + 1] - 1;
        for (int k = start; k < end; k++) {
            matrix->arcs[arc_idx].origin = j;
            matrix->arcs[arc_idx].dest = row_ind[k] - 1;
            matrix->arcs[arc_idx].value = values[k];
            arc_idx++;
        }
    }


    // Sort arcs by origin
    qsort(matrix->arcs, matrix->num_arcs, sizeof(Arc), compare_arcs);

    // Build row_ptr
    matrix->row_ptr = (int *)malloc((matrix->num_nodes + 1) * sizeof(int));
    int current_origin = 0;
    matrix->row_ptr[0] = 0;
    for (int k = 0; k < matrix->num_arcs; k++) {
        while (matrix->arcs[k].origin > current_origin) {
            current_origin++;
            matrix->row_ptr[current_origin] = k;
        }
    }
    for (int i = current_origin + 1; i <= matrix->num_nodes; i++) {
        matrix->row_ptr[i] = matrix->num_arcs;
    }

    free(col_ptr);
    free(row_ind);
    free(values);
    fclose(file);
    return matrix;
}
